/*=================
 *  CS8395 - HW2 - Part 2
 *  This starter code visits each function in the given program.
 *  Your job is to implement rudimentary stack canaries to protect against stack overflow attacks.
 =================*/

#include <set>
#include "InstrumentFunctions.h"

#include "llvm/IR/IRBuilder.h"
#include "llvm/Passes/PassPlugin.h"
#include "llvm/Passes/PassBuilder.h"
#include "llvm/IR/InstIterator.h"
#include "llvm/Support/raw_ostream.h"
#include "llvm/Transforms/Utils/BasicBlockUtils.h"
#include <cstdlib>


using namespace llvm;

#define DEBUG_TYPE "cs8395-hw2"




//-----------------------------------------------------------------------------
// The runOnModule method is invoked when LLVM encounters a "module", e.g., a compilation unit. 
// For the purposes of CS8395 HW2, this is equivalent to "a single .c file"
//-----------------------------------------------------------------------------
bool InstrumentFunctions::runOnModule(Module &M) {
  bool instrumentedAsLeastOnce = false;

  auto &CTX = M.getContext();
  PointerType *PrintfArgTy = PointerType::getUnqual(Type::getInt8Ty(CTX));

  // STEP 1: Inject the declaration of printf
  // This makes "printf" available to the program later.  Once it's
  // declare here, we can insert instrumentation that uses Printf
  // later. 
  // ----------------------------------------
  FunctionType *PrintfTy = FunctionType::get(
      IntegerType::getInt32Ty(CTX),
      PrintfArgTy,
      /*IsVarArgs=*/true);

  FunctionCallee Printf = M.getOrInsertFunction("printf", PrintfTy);

  // Set attributes as per inferLibFuncAttributes in BuildLibCalls.cpp
  Function *PrintfF = dyn_cast<Function>(Printf.getCallee());
  PrintfF->setDoesNotThrow();
  PrintfF->addParamAttr(0, Attribute::NoCapture);
  PrintfF->addParamAttr(0, Attribute::ReadOnly);


  // STEP 2: Inject a global variable that will hold the printf format string
  // For CS8395 HW2, this is just our alert that you can use that the
  // stack has been smashed. 
  // ------------------------------------------------------------------------
  llvm::Constant *PrintfFormatStr = llvm::ConstantDataArray::getString(CTX, "Function %s is being executed!\n");

  // we create a new global to hold the format string... it makes it
  // easier to refer to later when we need to insert a printf call.
  Constant *PrintfFormatStrVar = M.getOrInsertGlobal("PrintfFormatStr", PrintfFormatStr->getType());
  dyn_cast<GlobalVariable>(PrintfFormatStrVar)->setInitializer(PrintfFormatStr);

  // STEP 3: For each function in the module, insert instrumentation code
  // CS8395:   You only need to edit this loop!
  // ----------------------------------------------------------------
    for (auto &F : M) {
        // skip forward declarations
        if (F.isDeclaration())
            continue;

        // Get an IR builder. Sets the insertion point to the top of the function
        IRBuilder<> Builder(&*F.getEntryBlock().getFirstInsertionPt());
        
        // create a new stack canary for this function
        std::string CanaryName = "CanaryFor_" + std::string(F.getName());
        
        // initialize to some random value
        // for CS8395, I realize the RNG is not seeded.  This is not
        // meant to be an actual solution (stack canaries are already
        // implemented by modern compilers).  You can leave it
        // unseeded. 
        int canaryValue = rand();

        // insert allocation instructions in LLVM
        // this has the effect of introducing a new stack variable.
        // if an attacker smashes the stack, they will need to overwrite
        // the value we create here.
        AllocaInst* newStackCanary = Builder.CreateAlloca(
            IntegerType::getInt32Ty(CTX),
            nullptr,
            CanaryName);
        // Store instruction will put the value on the stack at runtime
        // In particular, we will store a randomly-generated value that we
        // will compare later!
        Builder.CreateStore(
            ConstantInt::get(CTX,APInt(32,canaryValue)),
            newStackCanary
        );



        // cs8395 start editing here
        // Your job is to change the given function (called F)
        // so that it contains instructions that check the value of
        // the canary that was created above _before_ the function
        // returns. 
        //
        // Consider looking for where the function ends (what type of
        // instruction signals the end of a fuction?), then load up
        // the canary value, and compare the value retrieved from the
        // stack with the constant value defined above 
        // (newStackCanary vs. canaryValue)


        // hint: you can build a worklist of particular instruction
        // types... 
        //
        //  for (inst_iterator I = inst_begin(F), E = inst_end(F); i != E; ++i)  ... 
        //
        // there's also an LLVM builtin called "isa" that tells you
        // whether a particular instruction is of a particular type...
        //
        // bool isAllocaInst = isa<AllocaInst>( &*I )





        // end cs8395 hw2


        // this code will insert a printf call...
            
            // get the format string we created way above.
            Value *FormatStrPtr = Builder.CreatePointerCast(PrintfFormatStrVar, PrintfArgTy, "formatStr");
            
            // We want to tell the user which function's stack is
            // smashed at runtime, so we also load up the function's
            // name. 
            auto FuncName = Builder.CreateGlobalStringPtr(F.getName());

            // Finally, we emit a call instruction to call Printf. 
            Builder.CreateCall(
                Printf, {FormatStrPtr, FuncName} 
            );




        // LLVM wants to know when we change part of the control flow
        // graph for subsequent optimization passes.  We play nice and
        // say we changed something...
        instrumentedAsLeastOnce = true;
    }

    return instrumentedAsLeastOnce;
}



PreservedAnalyses InstrumentFunctions::run(llvm::Module &M, llvm::ModuleAnalysisManager &) {
    bool Changed =  runOnModule(M);

    return (Changed ? llvm::PreservedAnalyses::none()
                    : llvm::PreservedAnalyses::all());
}

bool LegacyInstrumentFunctions::runOnModule(llvm::Module &M) {
    bool Changed = Impl.runOnModule(M);
    return Changed;
}

//-----------------------------------------------------------------------------
// New PM Registration  No need to touch this.
//-----------------------------------------------------------------------------
llvm::PassPluginLibraryInfo getInstrumentFunctionsPluginInfo() {
    return {LLVM_PLUGIN_API_VERSION, "cs8395-hw2", LLVM_VERSION_STRING,
          [](PassBuilder &PB) {
            PB.registerPipelineParsingCallback(
                [](StringRef Name, ModulePassManager &MPM,
                   ArrayRef<PassBuilder::PipelineElement>) {
                  if (Name == "cs8395-hw2") {
                    MPM.addPass(InstrumentFunctions());
                    return true;
                  }
                  return false;
                });
          }};
}

extern "C" LLVM_ATTRIBUTE_WEAK ::llvm::PassPluginLibraryInfo
llvmGetPassPluginInfo() {
  return getInstrumentFunctionsPluginInfo();
}

//-----------------------------------------------------------------------------
// Legacy PM Registration   Also no need to touch this
//-----------------------------------------------------------------------------
char LegacyInstrumentFunctions::ID = 0;

// Register the pass - required for (among others) opt
static RegisterPass<LegacyInstrumentFunctions>
    X(/*PassArg=*/"legacy-cs8395-hw2", /*Name=*/"LegacyInstrumentFunctions",
      /*CFGOnly=*/false, /*is_analysis=*/false);
